import logging
import threading

from kikyo import Kikyo, DataHub

from kikyopp.consumer.base import BaseConsumer

log = logging.getLogger(__name__)


class KikyoConsumer(BaseConsumer):

    def __init__(self, worker):
        self.worker = worker
        self.is_running = True
        self.consumer = None
        self._consumer_lock = threading.Lock()

    @property
    def kikyo(self) -> Kikyo:
        return self.worker.kikyo

    @property
    def worker_name(self) -> str:
        return self.worker.name

    def run(self, name):
        with self._consumer_lock:
            if self.consumer is None:
                datahub = self.kikyo.component(cls=DataHub)
                self.consumer = datahub.subscribe(
                    name,
                    subscription_name=f'kikyopp.{self.worker_name}',
                    auto_ack=False,
                )

        while self.is_running:
            try:
                with self._consumer_lock:
                    if not self.is_running:
                        break
                    msg = self.consumer.receive()

                data = msg.value
                self.worker.process(data)
                if not self.worker.debug:
                    self.consumer.ack(msg)
            except Exception as e:
                log.error(f'Error when consume data: {e}')

    def stop(self):
        self.is_running = False



YURI_API_HOST = 'http://yuri.app.kdsec.org'
TIKA_HOST = 'http://tika.app.kdsec.org'
PADDLEOCR_HOST = 'http://paddleocr.app.kdsec.org'
PDF_CONVERTER_HOST = 'http://pdf-converter.app.kdsec.org'
FILE_CONTENT_LIMIT = 5000



import base64
import gzip
import io
import mimetypes
import re
from typing import List, Optional

import fitz
import requests
from fitz import Pixmap
from pydantic import BaseModel
from requests_toolbelt import MultipartEncoder

from kikyopp.utils.constants import YURI_API_HOST, TIKA_HOST, PADDLEOCR_HOST, PDF_CONVERTER_HOST, FILE_CONTENT_LIMIT
from kikyopp.utils.retry import retry_rest_api


def pdf_to_png(data: bytes, limit: Optional[int] = None) -> List[bytes]:
    result = []
    pdf = fitz.Document(stream=io.BytesIO(data), filetype='pdf')
    for i, page in enumerate(pdf):
        # 将每一页pdf读取为图片
        img: Pixmap = page.getPixmap()
        result.append(img.tobytes())

        if limit is not None and len(result) >= limit:
            break
    return result


@retry_rest_api
def download_file_from_bus(file_id: str) -> bytes:
    resp = requests.get(
        f'{YURI_API_HOST}/filebus/download',
        params={
            'file_id': file_id,
        },
        timeout=20,
    )
    resp.raise_for_status()
    return resp.content


@retry_rest_api
def download_by_file_link(file_link: str) -> bytes:
    resp = requests.get(
        file_link,
        timeout=20,
    )
    resp.raise_for_status()
    return resp.content


@retry_rest_api
def extract_content(source: bytes) -> Optional[str]:
    resp = requests.put(
        f'{TIKA_HOST}/tika',
        data=source,
        timeout=20,
    )
    resp.raise_for_status()
    return resp.content.decode('utf-8')


@retry_rest_api
def extract_content_by_ocr(source: bytes) -> Optional[str]:
    image = base64.b64encode(source).decode('utf8')
    data = {"images": [image]}
    resp = requests.post(
        url=f'{PADDLEOCR_HOST}/predict/ocr_system',
        json=data,
        timeout=20,
    )
    resp.raise_for_status()
    res = resp.json()['results'][0]
    t = []
    for r in res:
        t.append(r['text'])
    return ''.join(t)


content_file_ext = {
    'doc': 'application/msword',
    'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    'pdf': 'application/pdf',
    'ppt': 'application/vnd.ms-powerpoint',
    'pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    'xls': 'application/vnd.ms-excel',
    'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
}


@retry_rest_api
def doc_to_pdf(source: bytes, filename: str = None) -> Optional[bytes]:
    data = None
    if filename:
        ext = filename.rsplit('.', maxsplit=1)[-1]
        if ext in content_file_ext:
            data = (f'data.{ext}', io.BytesIO(source), content_file_ext[ext])

    if data is None:
        return

    payload = MultipartEncoder({'data': data})
    resp = requests.post(
        f'{PDF_CONVERTER_HOST}/lool/convert-to/pdf',
        data=payload,
        headers={'Content-Type': payload.content_type},
        timeout=20,
    )
    resp.raise_for_status()
    return resp.content


@retry_rest_api
def extract_file_meta(source: bytes) -> dict:
    resp = requests.put(
        f'{TIKA_HOST}/meta',
        data=source,
        headers={'Accept': 'application/json'},
        timeout=20,
    )
    resp.raise_for_status()
    return resp.json()


class FileInspection(BaseModel):
    file_bytes: bytes = None
    file_ext: str = None
    content_type: str = None
    content_encoding: str = None


def inspect_file(file_bytes: bytes) -> FileInspection:
    meta = extract_file_meta(file_bytes)

    content_type = meta.get('Content-Type')
    if content_type:
        content_type = content_type.split(';')[0]

    if content_type == 'application/gzip':
        file_bytes = gzip.decompress(file_bytes)
        meta = extract_file_meta(file_bytes)
        content_type = meta.get('Content-Type')
        if content_type:
            content_type = content_type.split(';')[0]

    content_encoding = meta.get('Content-Encoding')
    if content_type is not None:
        file_ext: Optional[str] = mimetypes.guess_extension(content_type)
        if file_ext:
            file_ext = file_ext[1:]
    else:
        file_ext = None

    return FileInspection(
        file_bytes=file_bytes,
        file_ext=file_ext,
        content_type=content_type,
        content_encoding=content_encoding,
    )


blank_reg = re.compile(r'\s+')
word_blank_reg = re.compile(r'([\u4E00-\u9FA5])\s+([\u4E00-\u9FA5])')


def normalize_file_content(content: str) -> str:
    content = blank_reg.sub(' ', content)
    content = word_blank_reg.sub('', content)
    content = word_blank_reg.sub('', content)

    if len(content) > FILE_CONTENT_LIMIT:
        content = content[:FILE_CONTENT_LIMIT]
    return content



import inspect
import time
from functools import wraps

from requests import HTTPError

RETRY_HTTP_CODES = {500, 502, 503, 504, 522, 524, 408, 429}


class NotRetry(Exception):
    @property
    def error(self) -> str:
        return self.args[0]


def retry_rest_api(n: int = None, wait: int = 3):
    def wrapper(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            last_error = None
            k = 3 if n is None else n
            while k >= 0:
                try:
                    return func(*args, **kwargs)
                except NotRetry as e:
                    last_error = e.error
                    k = 0
                except HTTPError as e:
                    last_error = e
                    if e.response is not None and e.response.status_code not in RETRY_HTTP_CODES:
                        k = 0
                k -= 1
                if k >= 0 and wait > 0:
                    time.sleep(wait)
            if last_error is not None:
                raise last_error

        return wrap

    if inspect.isfunction(n):
        f = n
        n = None
        return wrapper(f)
    return wrapper



import inspect
import json
import logging
import os
import threading
from collections import defaultdict
from functools import partial

from kikyo import Kikyo, DataHub
from kikyo.settings import Settings

from kikyopp.annotation import KIKYOPP_SOURCE, KIKYOPP_SINK
from kikyopp.consumer.base import BaseConsumer
from kikyopp.consumer.kikyo import KikyoConsumer

log = logging.getLogger(__name__)


class BaseWorker:
    name: str
    kikyo: Kikyo
    consumer: BaseConsumer

    def __init__(self, settings: dict):
        self.settings = Settings(settings)

        log.info('settings: %s', json.dumps(self.settings, ensure_ascii=False, indent=4))

        self._init_env()
        assert self.name is not None
        self.debug = self.settings.get('debug', False)
        self.workers = self.settings.get('workers', 1)

        self._sink = None
        self._sink_method = None
        self._source = None
        self._source_method = None
        self._producer = None

        self._source = defaultdict(list)
        for name in dir(self):
            method = getattr(self, name)
            if inspect.ismethod(method):
                if hasattr(method, KIKYOPP_SOURCE):
                    self._source = getattr(method, KIKYOPP_SOURCE)
                    self._source_method = method
                if hasattr(method, KIKYOPP_SINK):
                    self._sink = getattr(method, KIKYOPP_SINK)
                    self._sink_method = method

        self.is_running = False

        self._init_components()
        if self._sink:
            self._producer = self.kikyo.component(cls=DataHub).create_producer(self._sink)

    def _init_env(self):
        worker_name = os.environ.get('KIKYOPP_WORKER_NAME')
        if worker_name:
            self.name = worker_name

    def _init_components(self):
        from kikyo import configure_by_consul

        self.kikyo = configure_by_consul(self.settings['kikyo_config_url'])
        self.consumer = KikyoConsumer(self)

    def start(self):
        if self.is_running:
            return

        if not self._source:
            log.warning('No defined source')

        self.is_running = True

        jobs = []
        for i in range(self.workers):
            t = threading.Thread(target=partial(self.consumer.run, self._source), daemon=True)
            t.start()
            jobs.append(t)

        for job in jobs:
            job.join()

    def process(self, data: dict):
        try:
            res = self._source_method(data)
            if res:
                if self._sink:
                    if not isinstance(res, list):
                        res = [res]
                    for r in res:
                        self._producer.send(r)
                        self._sink_method(r)
        except Exception as e:
            log.error(f'Error occurred: {e}', exc_info=True)

    def stop(self):
        if not self.is_running:
            return
        self.is_running = False
        self.consumer.stop()



import base64
import logging
from email.utils import getaddresses
from typing import Optional

import mailparser
from mailparser import MailParser

from kikyopp.utils.file import normalize_file_content
from kikyopp.utils.selector import Selector

log = logging.getLogger(__name__)


class BaseEmailParser:
    def parse_email(self, file_content: bytes) -> dict:
        data = {}
        mail = mailparser.parse_from_bytes(file_content)
        try:
            self._parse_headers(mail, data)
            self._parse_content(mail, data)
        except Exception as e:
            log.error('Parse email error: %s', e)
        return data

    def _parse_headers(self, mail: MailParser, data: dict):
        data['subject'] = mail.headers.get('Subject')
        data['mail_from'] = mail.headers.get('From')
        data['mail_to'] = mail.headers.get('To')
        data['mail_cc'] = mail.headers.get('Cc')
        data['reply_to'] = mail.headers.get('Reply-To')
        for i in ['mail_from', 'mail_to', 'mail_cc', 'reply_to']:
            data[i] = self.parse_address(data[i])

    def parse_address(self, address_str: str):
        if not address_str:
            return
        result = []
        s = getaddresses([address_str])
        for a in s:
            if a and a[1]:
                result.append(a[1])
        if len(result) == 0:
            return
        if len(result) == 1:
            return result[0]
        return result

    def _parse_content(self, mail: MailParser, data: dict):
        if 'content' not in data:
            data['content'] = ''
        if 'attachments' not in data:
            data['attachments'] = []

        for i in mail.text_plain:
            data['content'] += _read_index_content('plain', i)
        for i in mail.text_html:
            data['content'] += _read_index_content('html', i)

        for i in mail.attachments:
            content = base64.b64decode(i['payload'])
            filepath = self._save_attachment(i.get('filename'), i.get('mail_content_type'), content)
            data['attachments'].append({
                'filename': i.get('filename'),
                'filepath': filepath,
                'content_type': i.get('mail_content_type'),
            })

    def _save_attachment(self, filename: str, content_type: str, content: bytes) -> Optional[str]:
        pass


def _read_index_content(subtype, body) -> str:
    content = ''
    if subtype == 'html':
        content = _get_text_from_html(body)
        if content:
            content = normalize_file_content(content)
    elif subtype == 'plain':
        content = body
        if content:
            content = normalize_file_content(content)
    return content


def _get_text_from_html(html):
    try:
        selector = Selector(html)
        return selector.css('body').text[0]
    except Exception:
        return ''



import requests

from kikyopp.utils.constants import YURI_API_HOST
from kikyopp.utils.retry import retry_rest_api


@retry_rest_api
def predict_bm_domain(content: str) -> list:
    resp = requests.post(
        f'{YURI_API_HOST}/bm-domain/api/predict-bm-domain',
        json={'text': content},
        timeout=20,
    )
    resp.raise_for_status()
    return resp.json()['result']


@retry_rest_api
def extract_keywords(content: str) -> list:
    resp = requests.post(
        f'{YURI_API_HOST}/keywords/api/tag/keywords',
        json={'content': content},
        timeout=20,
    )
    resp.raise_for_status()
    return resp.json()['keywords']


@retry_rest_api
def predict_themes(content: str) -> list:
    resp = requests.post(
        f'{YURI_API_HOST}/theme/api/tag/themes',
        json={'content': content},
        timeout=20,
    )
    resp.raise_for_status()
    return resp.json()['result']
